import pandas as pd
import numpy as np
from skimage.morphology import binary_dilation, binary_erosion, dilation, erosion, disk
from skimage.measure import label, regionprops
from yeastvision.models.unet import unet
from yeastvision.track.cell import getBirthFrame, getLifeData
import matplotlib.pyplot as plt 
from yeastvision.track.utils import normalize_dict_by_sum
from tqdm import tqdm
from skimage.morphology import skeletonize
import matplotlib.cm as cm
import sys
import numpy as np
import matplotlib.pyplot as plt
from PyQt5.QtWidgets import QApplication, QMainWindow, QVBoxLayout, QWidget
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas



def count_objects(labeledMask):
    return np.count_nonzero(np.unique(labeledMask))

def get_daughters(tree, cell):
    potentialDaughters = tree[cell-1]
    daughters = list(np.where(potentialDaughters)[0]+1)
    return daughters

def has_daughters(tree, cell):
    return np.any(tree[cell-1,:])   

    
def get_generations(tree, initial_generation):
    generations = [[]]
    for val in initial_generation:
        generations[0].append(val)
        get_generations_recur(tree, 1, generations, 0, val)
    print(generations)
    return generations

def get_generations_recur(tree, generation, generations, max_generation, cell):
    if not has_daughters(tree, cell):
        return 
    
    daughters = get_daughters(tree, cell)
    if generation>=len(generations):
        generations.append([])
        max_generation = generation
    for daughter in daughters:
        generations[generation].append(daughter)
        get_generations_recur(tree, generation+1, generations, max_generation, daughter)

def get_gametes(skeleton, cyto, idx):
    found_gametes = False
    overlap = []
    previous_overlaps = []
    while (not found_gametes) and idx>=0:
        mask = cyto[idx]
        overlap = np.unique(mask[skeleton>0])
        overlap = overlap[overlap!=0]

        previous_overlaps.append(overlap)
        found_gametes = len(overlap)==2
        idx -=1
    
    if len(overlap)==0:
        for previous_overlap in previous_overlaps:
            if len(previous_overlap)>0:
                overlap = previous_overlap
                break

    if len(overlap)>2:
        overlap = overlap[0:2]
    else:
        overlap = np.append(overlap, [None, None])
    return found_gametes, overlap

def is_mating_cell(mat, cyto, cell):
    return np.all(mat[cyto==cell]==cell)

def compute_mating_lineage(tracked, cyto):

    cellVals = np.unique(cyto)
    cellVals = cellVals[cellVals!=0]

    matingCells = np.unique(tracked)
    matingCells = matingCells[matingCells!=0]

    numCells = np.count_nonzero(matingCells)
    
    print("computing lineages for", numCells, "cells")
    gamete_dict = {
                  "gamete1":[],
                    "gamete2":[],
                    "isMating":[],
                    "foundGametes":[]}

    for label in tqdm(cellVals):
        isMating = label in matingCells
        if isMating:
            birth = getBirthFrame(tracked, label)
            mat_birth_mask = (tracked[birth] == label).astype(np.uint8)
            found_gametes, gametes = get_gametes(skeletonize(mat_birth_mask),
                                                 cyto, birth-1)
        else:
            found_gametes = False
            gametes = [None, None]
        
        gamete_dict["gamete1"].append(gametes[0])
        gamete_dict["gamete2"].append(gametes[1])
        gamete_dict["isMating"].append(isMating)
        gamete_dict["foundGametes"].append(found_gametes)
    
    return pd.DataFrame.from_dict(gamete_dict)

class LineageConstruction():
    def __init__(self, segMasks, budMasks = None, backskip = 0, forwardskip = 0):
        self.cellMasks = segMasks 
        self.neckMasks = budMasks
        self.backskip = backskip
        self.forskip = forwardskip
        self.maxT = len(self.cellMasks)-1
        self.budNET = None
        
        self.labelMasks()

        # print("--shaving masks--")
        # self.shaveMasks()

        # self.labelMasks()
    

    
    def shaveMasks(self):
        i = 0
        for cellmask, neckmask in zip(self.cellMasks, self.neckMasks):
            print("\tshave mask", i)
            newNeckMask = np.zeros_like(neckmask)
            if np.any(neckmask):
                for neckI in np.unique(neckmask):
                    if neckI>0:
                        print("\t\t", neckI)
                        newNeckMask += self.processNeckMask(cellmask, neckmask==neckI)
            self.neckMasks[i] = newNeckMask      
            i+=1


    def processNeckMask(self, segIm, neckMask):
        processCount = 0
        overlapCount = count_objects(segIm[neckMask])
        r = 1
        newMask = neckMask.copy()
        footprint =  disk(r)
        while overlapCount != 2 and processCount<2:
            if overlapCount>2:
                newMask = binary_erosion(newMask, footprint = footprint)
            elif overlapCount<2:
                newMask = binary_dilation(newMask, footprint = footprint)
            overlapCount = count_objects(segIm[newMask])
            processCount+=1
        
        if overlapCount !=2:
            return neckMask.astype(np.uint8)
        else:
            return newMask
        
            
    def labelMasks(self):
        self.neckMasks = [label((mask>0).astype(np.uint16)) for mask in self.neckMasks]

    def getInitialCells(self):
        if count_objects(self.cellMasks[0])>1:
                self.initialCells = np.unique(self.neckMasks[0])
                self.initialCells = self.initialCells[self.initialCells!=0]

                if count_objects(self.neckMasks[0])>0:
                    pass
        else:
            self.initialCells = None

    def computeLineages(self):
        numCells = np.count_nonzero(np.unique(self.cellMasks))
        cellVals = np.unique(self.cellMasks)
        cellVals = cellVals[cellVals!=0]

        print("computing lineages for", numCells, "cells")
        daughterArray = np.zeros((numCells, numCells), dtype = bool) # rows = cell, column = daughter
        motherDict = getLifeData(self.cellMasks)
        motherDict["cell"] = cellVals
        motherDict["mother"] = []
        motherDict["confidence"] = []

        for cellId in tqdm(cellVals):
            mother, confidence = self.getMother(cellId)
            if mother:
                daughterArray[mother-1, cellId-1] = True
            motherDict["mother"].append(mother)
            motherDict["confidence"].append(confidence)
        
        motherDF = pd.DataFrame.from_dict(motherDict)
        #daughterDF = pd.DataFrame(data = daughterArray, index = cellVals, columns=cellVals)
        #motherDF = motherDF.set_index("cell")
        return daughterArray, motherDF[["birth", "death", "mother", "confidence"]]
    
    def getMother(self, cellNum):
        '''
        1) get bud that overlaps with daughter the most
        2) '''
        birth = getBirthFrame(self.cellMasks, cellNum)
        birthFrame = self.cellMasks[birth]
        eligibleFrames = slice(max(0, birth-self.backskip), min(self.maxT+1, birth+self.forskip))


        overlapcounts = {}
        i = 0
        for cellframe, budframe in zip(self.cellMasks[eligibleFrames], self.neckMasks[eligibleFrames]):
            if i < birth:
                cellframe = self.cellMasks[birth]

            potentialBuds, counts= np.unique(budframe[cellframe==cellNum], return_counts=True)
            if np.any(potentialBuds>0):
                potentialBuds, counts = potentialBuds[potentialBuds!=0], counts[potentialBuds!=0]
                correctBud = potentialBuds[np.argmax(counts)]
                mothers, counts = np.unique(cellframe[budframe==correctBud], return_counts=True)
                for mother, count in zip(mothers, counts):
                    if mother !=cellNum and mother!=0:
                        if mother in overlapcounts:
                            overlapcounts[mother] += count
                        else:
                            overlapcounts[mother] = count  
            i+=1   
        if overlapcounts:
            overlapcounts = normalize_dict_by_sum(overlapcounts)
            mother = max(overlapcounts, key = overlapcounts.get)
            confidence = overlapcounts[mother]

            if birth==0:
                motherArea = np.sum(birthFrame == mother)
                daughterArea = np.sum(birthFrame==cellNum)
                if daughterArea > motherArea:
                    return None, 0
            return mother, round(confidence,2)
        else:
            return None, 0














    



    
    









from yeastvision.models.cp import CustomCPWrapper

class SpoSeg(CustomCPWrapper):
    hyperparams  = { 
    "mean_diameter":30,
    "flow_threshold":0.4, 
    "cell_probability_threshold": 0.0}
    tpes = [float, float, float]
    def __init__(self, params, weights):
        super().__init__(params, weights)



import matplotlib.cm as cm
import sys
import numpy as np
import matplotlib.pyplot as plt
from PyQt5.QtWidgets import QApplication, QMainWindow, QVBoxLayout, QWidget
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

def plot_tree(matrix, selected_cells=[]):
    fig, ax = plt.subplots(figsize=(10, 6))
    y_positions = {}
    current_y = 0  # Starting position

    # Bright color list for selected cells
    bright_colors = [
        '#FF0000',  # Bright red
        '#00FF00',  # Bright green
        '#0000FF',  # Bright blue
        '#FF00FF',  # Bright magenta
        '#00FFFF',  # Bright cyan
        '#FFFF00',  # Bright yellow
    ]
    cell_colors = {selected_cells[i]: bright_colors[i % len(bright_colors)] for i in range(len(selected_cells))}

    def plot_lineage(mother, y):
        daughters = matrix[matrix[:, 3] == mother]
        daughters = daughters[daughters[:, 1].argsort()[::-1]]

        mother_color = cell_colors.get(mother, 'grey')

        for daughter in daughters:
            cell_index, birth_frame, death_frame, _ = daughter
            y += 1

            while y in y_positions.values():
                y += 1

            color = mother_color if cell_index in cell_colors or mother_color != 'grey' else 'grey'
            ax.plot([birth_frame, death_frame], [y, y], color=color)
            ax.text(death_frame + 1, y, str(cell_index), color=color, verticalalignment='center',
                    horizontalalignment='left')

            ax.plot([birth_frame, birth_frame], [y_positions[mother], y], color=mother_color)

            y_positions[cell_index] = y
            y = plot_lineage(cell_index, y)

        return y

    no_mothers = matrix[matrix[:, 3] == -1]
    for row in no_mothers:
        cell_index, birth_frame, death_frame, _ = row
        color = cell_colors.get(cell_index, 'grey')
        ax.plot([birth_frame, death_frame], [current_y, current_y], color=color)
        ax.text(death_frame + 1, current_y, str(cell_index), color=color, verticalalignment='center',
                horizontalalignment='left')
        y_positions[cell_index] = current_y
        current_y = plot_lineage(cell_index, current_y) + 1

    ax.set_ylim(-1, current_y + 1)
    ax.set_xlabel('Frame')
    ax.get_yaxis().set_visible(False)
    plt.grid(True)
    plt.tight_layout()
    return fig

class App(QMainWindow):
    def __init__(self, matrix, selected):
        super().__init__()
        self.left = 100
        self.top = 100
        self.width = 640
        self.height = 480
        self.initUI(matrix, selected)

    def initUI(self, matrix, selected):
        self.setWindowTitle('Lineage Tree Plot')
        self.setGeometry(self.left, self.top, self.width, self.height)
        m = plot_tree(matrix, selected_cells=selected)
        canvas = FigureCanvas(m)
        self.setCentralWidget(canvas)
        self.show()

if __name__ == '__main__':
    matrix = np.array([
        [1, 0, 50, -1],
        [2, 0, 48, -1],
        [3, 2, 51, 1],
        [4, 3, 45, 1],
        [5, 4, 52, 2],
        [6, 5, 53, 2],
        [7, 10, 54, 3],
        [8, 15, 50, 3],
        [9, 18, 46, 4],
        [10, 22, 49, 4]
    ])

    selected = [1, 3, 7]

    app = QApplication(sys.argv)
    ex = App(matrix, selected)
    sys.exit(app.exec_())



from PyQt5.QtWidgets import QApplication, QComboBox, QStyledItemDelegate, QStyle, QMessageBox, QWidget, QVBoxLayout, QListView, QPushButton
from PyQt5.QtCore import QRect
from PyQt5.QtGui import QPainter


class Item:
    def __init__(self, name):
        self.name = name

    def delete(self):
        print(f"Item '{self.name}' is deleted!")
        items.remove(self)


items = [Item("Item 1"), Item("Item 2"), Item("Item 3")]


class RemoveItemDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super(RemoveItemDelegate, self).__init__(parent)

    def paint(self, painter, option, index):
        super().paint(painter, option, index)
        
        # Only paint the close icon if there are more than one items in the combobox
        if self.parent().count() > 1:
            close_icon =  self.parent().style().standardIcon(QStyle.SP_DockWidgetCloseButton)
            rect = self.get_close_button_rect(option.rect)
            close_icon.paint(painter, rect)

    def get_close_button_rect(self, item_rect):
        icon_size = 16
        return QRect(item_rect.right() - icon_size - 5, item_rect.center().y() - icon_size // 2, icon_size, icon_size)


class CustomListView(QListView):
    def __init__(self, combo, parent=None):
        super(CustomListView, self).__init__(parent)
        self._combo = combo

    def mousePressEvent(self, event):
        combo = self._combo
        delegate = combo.itemDelegate()
        index_under_mouse = self.indexAt(event.pos())

        if combo.count() <= 1:  # If only one item left, don't process any further
            super().mousePressEvent(event)
            return

        if isinstance(delegate, RemoveItemDelegate) and delegate.get_close_button_rect(self.visualRect(index_under_mouse)).contains(event.pos()):
            reply = QMessageBox.question(self, "Confirmation", "Are you sure you want to remove this item?", QMessageBox.Yes | QMessageBox.No)
            if reply == QMessageBox.Yes:
                item_to_remove = items[index_under_mouse.row()]
                item_to_remove.delete()
                combo.removeItem(index_under_mouse.row())
        else:
            super().mousePressEvent(event)

class CustomComboBox(QComboBox):
    def __init__(self, parent=None):
        super(CustomComboBox, self).__init__(parent)
        self.setView(CustomListView(self))
        self.setItemDelegate(RemoveItemDelegate(self))

    def addNewItem(self, item):
        self.addItem(item.name)
        items.append(item)

app = QApplication([])

window = QWidget()
layout = QVBoxLayout(window)

combo = CustomComboBox(window)
for item in items:
    combo.addItem(item.name)

layout.addWidget(combo)

# Button to add a new item
add_button = QPushButton("Add Item", window)
add_button.clicked.connect(lambda: combo.addNewItem(Item(f"Item {len(items) + 1}")))
layout.addWidget(add_button)

window.setLayout(layout)
window.resize(400, 300)
window.show()

app.exec_()




# from PyQt5.QtWidgets import QApplication, QComboBox, QStyledItemDelegate, QStyle, QMessageBox, QWidget, QVBoxLayout, QListView
# from PyQt5.QtCore import QRect
# from PyQt5.QtGui import QPainter

# class RemoveItemDelegate(QStyledItemDelegate):
#     def __init__(self, parent=None):
#         super(RemoveItemDelegate, self).__init__(parent)

#     def paint(self, painter, option, index):
#         super().paint(painter, option, index)
#         close_icon = self.parent().style().standardIcon(QStyle.SP_DockWidgetCloseButton)
#         rect = self.get_close_button_rect(option.rect)
#         close_icon.paint(painter, rect)

#     def get_close_button_rect(self, item_rect):
#         icon_size = 16
#         return QRect(item_rect.right() - icon_size - 5, item_rect.center().y() - icon_size // 2, icon_size, icon_size)

# class CustomListView(QListView):
#     def __init__(self, combo, parent=None):
#         super(CustomListView, self).__init__(parent)
#         self._combo = combo

#     def mousePressEvent(self, event):
#         combo = self._combo
#         delegate = combo.itemDelegate()
#         index_under_mouse = self.indexAt(event.pos())
        
#         if isinstance(delegate, RemoveItemDelegate) and delegate.get_close_button_rect(self.visualRect(index_under_mouse)).contains(event.pos()):
#             reply = QMessageBox.question(self, "Confirmation", "Are you sure you want to remove this item?",
#                                          QMessageBox.Yes | QMessageBox.No)
#             if reply == QMessageBox.Yes:
#                 combo.removeItem(index_under_mouse.row())
#         else:
#             super().mousePressEvent(event)

# app = QApplication([])

# window = QWidget()
# layout = QVBoxLayout(window)
# combo = QComboBox(window)
# combo.setView(CustomListView(combo))
# combo.setItemDelegate(RemoveItemDelegate(combo))
# combo.addItem("Item 1")
# combo.addItem("Item 2")
# combo.addItem("Item 3")
# layout.addWidget(combo)
# window.setLayout(layout)

# window.resize(400, 200)
# window.show()

# app.exec_()
